import platform
from pathlib import Path

import click
from mlflow_backend_utils.sdk import build_triton_stub as _build_triton_stub

from tsbk import TSBK_DIR, TSBK_K8S_SERVICE_ACCOUNT, TSBK_S3_PREFIX


def build_triton_stub(
    python_version: str,
    triton_version: str,
) -> Path:
    """Builds a conda environment given the requirements and packages it up with conda-pack and writes
    the resulting archive to the output_dir

    Args:
        python_version: the python version to use for the environment
        triton_version: the version of Triton to use for the stub

    Returns:
        path to the output stub
    """
    arch = platform.machine()

    output_dir = TSBK_DIR.joinpath("triton_python_stubs")
    output_dir.mkdir(parents=True, exist_ok=True)
    output_path = output_dir.joinpath(f"triton_python_backend_stub-{triton_version}-{arch}-{python_version}")

    if output_path.exists():
        return output_path

    if TSBK_S3_PREFIX:
        s3_path = (
            f"{TSBK_S3_PREFIX}/triton_python_stubs/triton_python_backend_stub-{triton_version}-{arch}-{python_version}"
        )
    else:
        s3_path = None

    click.secho(f"Building Triton Python Stub for {triton_version}/{python_version} on {arch}", fg="blue")
    _build_triton_stub(
        python_version=python_version,
        triton_version=triton_version,
        platform=arch,
        output_path=output_path,
        k8s_shared_s3_path=s3_path,
        k8s_service_account=TSBK_K8S_SERVICE_ACCOUNT,
    )

    return output_path



import numpy as np
from tritonclient.utils import np_to_triton_dtype


def parse_mlflow_signature_value(value: dict) -> tuple[str | None, str, list[int], int]:
    """Parses a MLFlow signature value into a tuple of (name, type, shape, batch size)

    Args:
        value: MLFlow signature value

    Returns:
        Tuple of (name, type, shape, batch size)
    """
    if value["type"] == "tensor":
        dtype = np_to_triton_dtype(np.dtype(value["tensor-spec"]["dtype"]))
        dtype = "STRING" if dtype == "BYTES" else dtype
        shape = value["tensor-spec"]["shape"]
        if len(shape) > 1:
            dims = shape[1:]
            batch_size = shape[0]
        else:
            dims = shape
            batch_size = 0
    elif value["type"] == "array":
        dtype = "STRING"
        dims = [-1]
        batch_size = 0
    elif value["type"] == "integer":
        dtype = "INT32"
        dims = [-1]
        batch_size = 0
    else:
        dtype = "STRING" if value["type"] == "string" else np_to_triton_dtype(np.dtype(value["type"]))
        dims = [-1]
        batch_size = 0

    return value.get("name"), f"{dtype}", dims, batch_size


def parse_mlflow_signature(mlflow_signature: tuple[list, list]) -> tuple[int, dict, dict]:
    """Parses an MLFlow model signature into Triton input and output schemas

    Args:
        mlflow_signature: MLFlow model signature

    Returns:
        Tuple of max batch size, input schema, and output schema
    """
    batch_sizes = set()
    inputs = dict()
    for i, model_input in enumerate(mlflow_signature[0]):
        name, dtype, dims, batch_size = parse_mlflow_signature_value(model_input)
        if name is None:
            name = f"input{i}"
        inputs[name] = dtype, dims, batch_size
        batch_sizes.add(batch_size)

    outputs = dict()
    for i, model_output in enumerate(mlflow_signature[1]):
        name, dtype, dims, batch_size = parse_mlflow_signature_value(model_output)
        if name is None:
            name = f"output{i}"
        outputs[name] = dtype, dims, batch_size
        batch_sizes.add(batch_size)

    if batch_sizes == {-1, 0}:
        for name, (dtype, dims, batch_size) in inputs.items():
            if batch_size == -1:
                inputs[name] = dtype, [-1] + dims, 0
        for name, (dtype, dims, batch_size) in outputs.items():
            if batch_size == -1:
                outputs[name] = dtype, [-1] + dims, 0
        batch_sizes.remove(-1)

    if len(batch_sizes) > 1:
        if 0 in batch_sizes:
            raise ValueError(f"Inconsistent batch sizes: {batch_sizes}")
        if -1 in batch_sizes:
            batch_sizes.remove(-1)
        batch_size = min(batch_sizes)
        for name, (dtype, dims, _) in inputs.items():
            inputs[name] = dtype, dims, batch_size
        for name, (dtype, dims, _) in outputs.items():
            outputs[name] = dtype, dims, batch_size
        batch_sizes = {batch_size}

    return batch_sizes.pop(), inputs, outputs


def python_version_for_triton_version(triton_version: str) -> str:
    """Returns the default Python version for a given Triton version.

    Args:
        triton_version: The Triton version.

    Returns:
        The default Python version for the given Triton version.
    """
    version_no = float(triton_version)
    if version_no >= 24.11:
        return "3.12"
    else:
        return "3.10"



import os.path
import shutil
import subprocess
import threading
import time
from pathlib import Path

import click
import mlflow_backend
import requests

from tsbk import DEFAULT_TRITON_VERSION, TSBK_DIR
from tsbk.model import TritonModel
from tsbk.test_case import TestResult
from tsbk.testing import TritonModelRepoTestPlan


class TritonModelRepo:
    def __init__(
        self,
        name: str,
        path: str | Path,
        models: dict[str, TritonModel | dict],
        triton_image: str = "nvcr.io/nvidia/tritonserver",
        triton_image_tag: str = f"{DEFAULT_TRITON_VERSION}-py3",
    ):
        """A Triton Model Repository.

        Args:
            name: The name of the model repository, which is used as the directory name in the Triton model repository.
            models: A dict of Triton models to deploy in the model repository.
            triton_image: The Docker image to use for the Triton server.
            triton_image_tag: The Docker image tag to use for the Triton server.
        """
        self.name = name
        self.models = models
        self.triton_image = triton_image
        self.triton_image_tag = triton_image_tag
        self.triton_version = self.parse_triton_version(self.triton_image_tag)

        self.models = {
            name: TritonModel(**model) if isinstance(model, dict) else model for name, model in self.models.items()
        }

        self.path = Path(path).absolute()
        for name, model in self.models.items():
            model.init(name=name, repo_path=self.path, triton_version=self.triton_version)

        self._resolve_ensemble_models()

        self.container_id = None
        self.http_url = None
        self.grpc_url = None
        self.metrics_url = None
        self.log_thread = None
        self.mlflow_backend_path = None

    @staticmethod
    def parse_triton_version(triton_image_tag: str):
        triton_version = triton_image_tag.split("-")[0] if "-" in triton_image_tag else triton_image_tag
        try:
            major, minor = triton_version.split(".")
            _ = int(major), int(minor)
            return f"{major}.{minor}"
        except ValueError:
            raise ValueError(
                f"Could not parse Triton version from image tag: {triton_image_tag}. Expected format is 'major.minor'"
            )

    def _resolve_ensemble_models(self):
        """Resolves ensemble models by ensuring all required models are included in the repo."""
        for model_name, model in self.models.items():
            if model.ensemble_models:
                for i, (ens_model_name, ens_model_version) in enumerate(model.ensemble_models):
                    if not any(_m == ens_model_name for _m in self.models.keys()):
                        raise ValueError(
                            f"Ensemble model {ens_model_name}:{ens_model_version} not found in the model repository."
                        )
                    if ens_model_version == -1:
                        self.models[model_name].ensemble_models[i] = (
                            ens_model_name,
                            self.models[ens_model_name].versions[-1].version,
                        )
                    elif not any(_mv.version == ens_model_version for _mv in self.models[ens_model_name].versions):
                        raise ValueError(
                            f"Ensemble model {ens_model_name}:{ens_model_version} not found in the model repository."
                        )

        click.secho("All ensemble models resolved successfully.", fg="green")

    @property
    def uses_mlflow_backend(self):
        return any([model.backend == "mlflow" for model in self.models.values()])

    def build(self):
        """Prepares all the models in this repo for serving."""
        if self.path.exists():
            click.secho("Warning: model repo already exists and will be overwritten", fg="yellow")
        shutil.rmtree(self.path, ignore_errors=True)
        click.secho("beginning model preparation...", fg="blue")
        for n, model in enumerate(self.models.values()):
            click.secho(f"preparing model {n+1} of {len(self.models)}: {model.name}", fg="blue")
            model.build()

        if self.uses_mlflow_backend:
            backend_dir = Path(mlflow_backend.__file__).parent.absolute().as_posix()
            backend_version = mlflow_backend.__version__
            self.mlflow_backend_path = TSBK_DIR.joinpath("backends", f"mlflow-{backend_version}")
            if not self.mlflow_backend_path.exists():
                click.secho(f"Copying mlflow backend to {self.mlflow_backend_path}", fg="blue")
                shutil.copytree(backend_dir, self.mlflow_backend_path, dirs_exist_ok=True)

    def run(
        self,
        http_port: int = 8000,
        grpc_port: int = 8001,
        metrics_port: int = 8002,
        shm_size: str = "4G",
        gpus: bool = False,
        env_file: str | None = None,
        detach: bool = False,
    ):
        click.secho("Running tritonserver in container...", fg="blue")
        http_port = http_port or ""
        grpc_port = grpc_port or ""
        metrics_port = metrics_port or ""

        mount_path = os.path.commonpath([TSBK_DIR.resolve(), self.path.resolve()])
        cmd = [
            "docker",
            "run",
            "-v",
            f"{mount_path}:/data",
            "-e",
            "MLFLOW_TRACKING_URI=file:///root/blackhole/mlruns",
            "--shm-size",
            shm_size,
        ]
        if self.uses_mlflow_backend:
            cmd.extend(["-v", f"{self.mlflow_backend_path.absolute().as_posix()}:/opt/tritonserver/backends/mlflow"])

        for host_port, target_port in [(http_port, 8000), (grpc_port, 8001), (metrics_port, 8002)]:
            cmd.extend([f"-p{host_port}:{target_port}"])

        if gpus:
            cmd.extend(["--gpus", "all"])

        if env_file:
            cmd.extend(["--env-file", env_file])

        if detach:
            cmd.append("-d")

        cmd.extend([f"{self.triton_image}:{self.triton_image_tag}"])

        cmd.extend(
            [
                f"tritonserver",
                f"--model-repository=/data/{self.path.relative_to(mount_path)}",
                "--strict-readiness",
                "true",
            ]
        )

        click.secho(f"Running command:\n{' '.join(cmd)}", fg="blue")

        result = subprocess.run(cmd, check=False, capture_output=detach)
        if result.returncode != 0:
            if detach:
                click.secho(f"Error running tritonserver: {result.stderr.decode()}", fg="red")
            raise RuntimeError("Error running tritonserver")

        if detach:
            click.secho("Triton server is running in detached mode.", fg="green")
            self.container_id = result.stdout.decode().strip()
            self._follow_container_logs()
            self.http_url = f"http://localhost:{http_port}"
            self.grpc_url = f"http://localhost:{grpc_port}"
            self.metrics_url = f"http://localhost:{metrics_port}/metrics"
            try:
                self._wait_for_server_ready(self.http_url)
            except KeyboardInterrupt:
                click.secho("Interrupted while waiting for server to be ready. Stopping container...", fg="red")
                self.stop()
                raise

    def _wait_for_server_ready(self, url: str, timeout: int = 60 * 5):
        """Waits for the Triton server to be ready by checking the health endpoint."""
        start_time = time.time()
        while True:
            try:
                response = requests.get(f"{url}/v2/health/ready")
                if response.status_code == 200:
                    click.secho("Triton server is ready!", fg="green")
                    return
            except requests.RequestException as e:
                click.secho(f"Waiting for Triton server to be ready: {e}", fg="yellow")
            if not self._container_still_running():
                raise RuntimeError("Triton server container has stopped unexpectedly.")
            if time.time() - start_time > timeout:
                raise TimeoutError(f"Triton server did not become ready within {timeout} seconds.")
            time.sleep(5)

    def _container_still_running(self):
        """Checks if the Triton server container is still running."""
        if not self.container_id:
            return False
        try:
            result = subprocess.run(
                ["docker", "inspect", "--format", "{{.State.Running}}", self.container_id],
                check=True,
                capture_output=True,
                text=True,
            )
            return result.stdout.strip() == "true"
        except subprocess.CalledProcessError:
            return False

    def _follow_container_logs(self):
        """Starts a new subprocess to follow and print container logs."""
        if not self.container_id:
            click.secho("No Triton server container is currently running.", fg="yellow")
            return

        def follow_logs():
            try:
                subprocess.run(["docker", "logs", "-f", self.container_id], check=True)
            except subprocess.CalledProcessError:
                click.secho("Error following container logs.", fg="red")

        self.log_thread = threading.Thread(target=follow_logs, daemon=True)
        self.log_thread.start()

    def test(
        self, url: str, ca_certs: str | None = None, headers: dict | None = None, grpc: bool = False
    ) -> list[TestResult]:
        """Runs the test cases defined for all models in this repository against the Triton server.

        Args:
            url: The URL of the Triton server to test against.
            ca_certs: Path to CA certificates for secure connections, if applicable.
            headers: Additional headers to include in the requests.
            grpc: Whether to use gRPC for testing.

        Returns:
            A list of TestResult objects containing the results of the tests.
        """
        test_plan = self.create_test_plan()
        return test_plan.run_tests(url=url, ca_certs=ca_certs, headers=headers, grpc=grpc)

    def create_test_plan(self) -> TritonModelRepoTestPlan:
        return TritonModelRepoTestPlan(
            model_plans=[model.create_test_plan() for model in self.models.values()],
            repo_models=self.repo_models,
        )

    @property
    def repo_models(self) -> list[tuple[str, int]]:
        """Returns a list of tuples of model name and version for all models in the repo."""
        repo_models = list()
        for model in self.models.values():
            for version in model.versions:
                repo_models.append((model.name, version.version))
        return repo_models

    def stop(self):
        """Stops the Triton server container if it is running."""
        if self.container_id:
            click.secho(f"Stopping Triton server container {self.container_id}...", fg="blue")
            subprocess.run(["docker", "stop", self.container_id], check=True)
            subprocess.run(["docker", "rm", self.container_id], check=True)
            self.container_id = None
        else:
            click.secho("No Triton server container is currently running.", fg="yellow")



import shutil
from copy import deepcopy
from functools import partial
from pathlib import Path
from typing import Optional
from urllib.parse import urlparse

from tsbk.test_case import TestCase
from tsbk.testing import TritonModelVersionTestPlan
from tsbk.utils import link_or_copy
from tsbk.utils.dbx import download_mlflow_model, get_input_example_from_model
from tsbk.utils.s3 import download_s3_path, s3_path_exists


class TritonModelVersion:
    def __init__(
        self,
        artifact_uri: str | None = None,
        python_model_file: str | None = None,
        version: int | None = None,
        test_cases: list[TestCase | dict] | None = None,
    ):
        """A Triton model version.

        Args:
            artifact_uri: The URI of the model artifact, which can be an MLflow model or an S3 object.
            python_model_file: The path to the Python model file, which is required for Python models.
            version: The version number of the model.
        """
        self.artifact_uri = artifact_uri
        self.python_model_file = python_model_file
        self.version = version
        self.test_cases = test_cases or []
        self.test_cases = [
            TestCase(**test_case) if isinstance(test_case, dict) else test_case for test_case in self.test_cases
        ]
        self._check_version_number()

        self.initialized: bool = False
        self.name: str | None = None
        self.path: Path | None = None
        self.backend: str | None = None
        self.platform: str | None = None
        self.flavor: str | None = None
        self.decoupled_transaction_policy: bool | None = None

    def __str__(self):
        return f"TritonModelVersion(artifact_uri={self.artifact_uri}, python_model_file={self.python_model_file}, version={self.version})"

    def _check_version_number(self):
        """Check that version is greater than 0"""
        if self.version is not None and self.version < 1:
            raise ValueError("version must be greater than 0")

    def init(
        self,
        name: str,
        path: Path,
        backend: str | None,
        platform: str | None,
        flavor: str | None,
        model_tests: list[TestCase],
        decoupled_transaction_policy: bool,
    ) -> None:
        """Initialize the model version with the given parameters."""
        self.name = name
        self.path = path
        self.backend = backend
        self.platform = platform
        self.flavor = flavor
        self.decoupled_transaction_policy = decoupled_transaction_policy
        for test in model_tests:
            self.test_cases.append(deepcopy(test))

        if self.artifact_uri and self.artifact_uri.startswith("models:/"):
            input_example = get_input_example_from_model(self.artifact_uri)
            if input_example is not None:
                self.test_cases.append(TestCase(inputs=input_example, expected_outputs={}))

        self._check_backend_or_platform()
        self._check_artifact_uri_or_python_model_file()
        self._check_python_model_file()
        self._check_artifact_uri()
        self.initialized = True

    def _check_backend_or_platform(self):
        if self.backend is None and self.platform is None:
            raise ValueError("Either backend or platform must be specified")
        if self.backend is not None and self.platform is not None:
            raise ValueError("Only one of backend or platform can be specified")

    def _check_artifact_uri_or_python_model_file(self):
        """Check that either artifact_uri or python_model_file is provided"""
        if self.platform != "ensemble" and (self.artifact_uri is None and self.python_model_file is None):
            raise ValueError("Either artifact_uri or python_model_file must be provided")

    def _check_python_model_file(self):
        if self.backend == "python" and not self.python_model_file and self.artifact_source[0] != "mlflow":
            raise ValueError("python_model_file must be specified for Python models that are not MLflow models")

    def _check_artifact_uri(self):
        if self.artifact_uri:
            uri = urlparse(self.artifact_uri)
            if uri.scheme not in ["models", "s3"]:
                raise ValueError(f"artifact_uri must be either an MLflow or S3 object, got {uri.scheme}")
            if uri.scheme == "s3" and not s3_path_exists(self.artifact_uri):
                raise ValueError(f"artifact_uri {self.artifact_uri} does not exist")

    def build(self) -> None:
        """Takes a configuration block defining a specific model version and handles the copying and organization of model assets in the version directory in a Triton-compatible format."""
        self.path.mkdir(parents=True)
        if self.artifact_uri:
            source, func = self.artifact_source
            copy_func = partial(func, origin_path=self.artifact_uri)

            match self.backend:
                case "mlflow":
                    copy_func(dst_path=self.path)
                case "python":
                    if source == "mlflow":
                        output_file_path = self.path.joinpath("mlflow_model").as_posix()
                        copy_func(dst_path=output_file_path)
                    else:
                        output_file_path = self.path.joinpath(self.artifact_uri.split("/")[-1])
                        copy_func(dst_path=output_file_path)

                case "pytorch":
                    if source == "mlflow":
                        model_path = copy_func()
                        # rel_path = os.path.relpath(model_path.joinpath("data/model.pth"), self.path)
                        link_or_copy(
                            model_path.joinpath("data/model.pth"),
                            self.path.joinpath("model.pt"),
                        )
                    else:
                        output_file_path = self.path.joinpath("model.pt").as_posix()
                        copy_func(dst_path=output_file_path)

                case "onnxruntime":
                    if source == "mlflow":
                        self.path.rmdir()
                        copy_func(dst_path=self.path)
                    else:
                        output_file_path = self.path.joinpath("model.onnx").as_posix()
                        copy_func(dst_path=output_file_path)

                case "tensorrt":
                    output_file_path = self.path.joinpath("model.plan").as_posix()
                    copy_func(dst_path=output_file_path)

                case _:
                    output_file_path = self.path.joinpath(self.artifact_uri.split("/")[-1])
                    copy_func(dst_path=output_file_path)

        if self.python_model_file:
            # Note: for now we expect that python_model_file to be built into container at WORKING_DIR/models/ path
            shutil.copy(
                self.python_model_file,
                self.path.joinpath("model.py"),
            )

        # ensemble models have no version assets, but need a dummy model version folder to be recognized by Triton
        if self.platform == "ensemble":
            keep_path = self.path.joinpath(".keep")
            keep_path.parent.mkdir(exist_ok=True)
            with open(keep_path, "w") as fp:
                fp.write("Need this model version folder for ensemble model.")

    @property
    def artifact_source(self) -> tuple[Optional[str], Optional[callable]]:
        """
        Determines the source of a given artifact_uri (S3 or MLFlow) and provides the appropriate function to copy the artifact from source to version folder.

        Returns:
            source: "mlflow" or "S3"
            copy_function: Function to use to copy the artifact from its origin to the Triton model version folder.
            profile: Profile to use if artifact is a DBX object. If the artifact is an S3 object, this is None
        """
        if not self.artifact_uri:
            return None, None
        elif self.artifact_uri.startswith("models:/"):
            return "mlflow", download_mlflow_model
        elif self.artifact_uri.startswith("s3://"):
            return "s3", download_s3_path
        raise ValueError(f"Invalid artifact_uri: {self.artifact_uri}")

    def create_test_plan(self) -> TritonModelVersionTestPlan:
        """Create a test plan for this model version."""
        if not self.initialized:
            raise ValueError("Model version must be initialized before creating a test plan")
        return TritonModelVersionTestPlan(
            name=self.name,
            version=self.version,
            test_cases=self.test_cases,
            decoupled_transaction_policy=self.decoupled_transaction_policy,
        )



import traceback
from dataclasses import dataclass

import numpy as np

from tsbk.triton_grpc_model_client import TritonGrpcModelClient
from tsbk.triton_http_model_client import TritonHttpModelClient
from tsbk.types import NumpyArray


@dataclass
class TestResult:
    """A class to hold the result of a test case execution."""

    success: bool
    model_name: str
    model_version: int
    message: str | None = None
    exception: BaseException | None = None
    traceback: str | None = None
    missing_test: bool = False
    decoupled_with_http: bool = False


class TestCase:
    def __init__(
        self,
        inputs: dict[str, np.ndarray | NumpyArray | dict],
        expected_outputs: dict[str, np.ndarray | NumpyArray | dict],
        allow_nan: bool = False,
        allow_inf: bool = False,
        rtol: float = 1.0e-5,
        atol: float = 1.0e-8,
    ):
        self.inputs = inputs
        for k, v in self.inputs.items():
            if isinstance(v, NumpyArray):
                self.inputs[k] = v.to_numpy()
            elif isinstance(v, dict):
                self.inputs[k] = NumpyArray(**v).to_numpy()

        self.expected_outputs = expected_outputs
        for k, v in self.expected_outputs.items():
            if isinstance(v, NumpyArray):
                self.expected_outputs[k] = v.to_numpy()
            elif isinstance(v, dict):
                self.expected_outputs[k] = NumpyArray(**v).to_numpy()

        self.allow_nan = allow_nan
        self.allow_inf = allow_inf
        self.rtol = rtol
        self.atol = atol

    def run(self, model_client: TritonHttpModelClient | TritonGrpcModelClient, decoupled: bool = False) -> TestResult:
        try:
            if decoupled:
                assert isinstance(
                    model_client, TritonGrpcModelClient
                ), "Decoupled inference only supported with GRPC client"
                outputs = model_client.infer_decoupled(self.inputs)
            else:
                outputs = model_client.infer(self.inputs)
            if set(self.expected_outputs.keys()) > set(outputs.keys()):
                return TestResult(
                    success=False,
                    model_name=model_client.model_name,
                    model_version=model_client.model_version,
                    message=f"Expected outputs {self.expected_outputs.keys()} not found in model outputs {outputs.keys()}",
                )
            for tensor_name, output in outputs.items():
                expected = self.expected_outputs.get(tensor_name)
                self.validate_output(output, expected, tensor_name, self.allow_nan, self.allow_inf, self.rtol, self.atol)
        except Exception as e:
            return TestResult(
                success=False,
                model_name=model_client.model_name,
                model_version=model_client.model_version,
                message=f"Inference failed for {model_client.model_name}:{model_client.model_version}",
                exception=e,
                traceback=traceback.format_exc(),
            )
        return TestResult(
            success=True,
            model_name=model_client.model_name,
            model_version=model_client.model_version,
            message=f"Inference succeeded for {model_client.model_name}:{model_client.model_version}",
        )

    @staticmethod
    def validate_output(
        output: np.ndarray,
        expected: np.ndarray | None,
        tensor_name: str,
        allow_nan: bool = False,
        allow_inf: bool = False,
        rtol: float = 1.0e-5,
        atol: float = 1.0e-8,
    ):
        """Standard function for validating model outputs. Raises a ValueError if the output is not as expected.

        Args:
            output: The model output to validate
            expected: The expected output. If None, only NaNs and Infs will be checked in the output
            tensor_name: The name of the tensor being validated
            allow_nan: Whether to allow NaNs in the output
            allow_inf: Whether to allow Infs in the output
            rtol: The relative tolerance to use for comparing floating point values
            atol: The absolute tolerance to use for comparing floating point values

        Raises:
            ValueError: If the output is not as expected
        """

        if output.dtype != object and np.isnan(output).any() and not allow_nan:
            raise ValueError(f"Output {tensor_name} contains NaNs")
        if output.dtype != object and np.isinf(output).any() and not allow_inf:
            raise ValueError(f"Output {tensor_name} contains Infs")

        if expected is not None:
            if output.shape != expected.shape:
                raise ValueError(
                    f"Output {tensor_name} has incorrect shape. Expected {expected.shape}, got {output.shape}"
                )

            if output.dtype != expected.dtype:
                raise ValueError(
                    f"Output {tensor_name} has incorrect dtype. Expected {expected.dtype}, got {output.dtype}"
                )

            if output.dtype == object:
                if not np.all(output == expected):
                    raise ValueError(f"Output {tensor_name} has incorrect values. Expected {expected}, got {output}")
            else:
                if not np.allclose(output, expected, rtol=rtol, atol=atol, equal_nan=allow_nan):
                    raise ValueError(f"Output {tensor_name} has incorrect values. Difference: {output - expected}")

    def to_dict(self) -> dict:
        """Convert the test case to a dictionary representation."""
        return {
            "inputs": {k: NumpyArray.from_numpy(v).to_dict() for k, v in self.inputs.items()},
            "expected_outputs": {k: NumpyArray.from_numpy(v).to_dict() for k, v in self.expected_outputs.items()},
            "allow_nan": self.allow_nan,
            "allow_inf": self.allow_inf,
            "rtol": self.rtol,
            "atol": self.atol,
        }

    @classmethod
    def from_dict(cls, d: dict) -> "TestCase":
        """Create a TestCase instance from a dictionary representation."""
        return cls(
            inputs=d["inputs"],
            expected_outputs=d["expected_outputs"],
            allow_nan=d.get("allow_nan", False),
            allow_inf=d.get("allow_inf", False),
            rtol=d.get("rtol", 1.0e-5),
            atol=d.get("atol", 1.0e-8),
        )



import os
import platform
import shutil
from pathlib import Path

import click


def get_platform_desc() -> str:
    """Returns the description for the current platform"""
    return f"{platform.system()}-{platform.machine()}"


def parse_requirements_file(requirements_path: Path) -> set[str]:
    """Parses all the requirements listed in the requirements file"""
    with open(requirements_path, "r") as fp:
        reqs = fp.readlines()
        reqs = [r.strip() for r in reqs]
        return set(reqs) - {""}


def parse_python_version(python_version: str) -> str:
    """Parses the passed python version ensures it takes the form of major.minor version (i.e. removing any patch version provided)"""
    try:
        major, minor = python_version.split(".")
        return f"{major}.{minor}"
    except ValueError:
        major, minor, _ = python_version.split(".")
        return f"{major}.{minor}"


def link_or_copy(src, dst):
    try:
        os.link(src, dst)
    except (PermissionError, OSError) as e:
        # If hard linking fails due to permission issues, fall back to copying
        click.secho(f"Hard linking failed with exception {e}, falling back to copying.", fg="yellow")
        shutil.copy2(src, dst)



import shutil
from pathlib import Path
from typing import Any

import click
import yaml
from mlflow_backend_utils.sdk import build_config
from mlflow_backend_utils.utils import parse_mlflow_signature

from tsbk.model_version import TritonModelVersion
from tsbk.test_case import TestCase
from tsbk.testing import TritonModelTestPlan
from tsbk.types import TritonDType
from tsbk.utils import link_or_copy, parse_python_version
from tsbk.utils.conda import build_conda_env
from tsbk.utils.dbx import (
    download_mlflow_model,
    find_matching_dbx_models,
    get_flavor_for_models,
    get_mlflow_model_requirements,
    get_python_version_for_mlflow_models,
)
from tsbk.utils.pbtxt import parse_pbtxt
from tsbk.utils.stubs import build_triton_stub
from tsbk.utils.triton import python_version_for_triton_version

BACKENDS_EXCLUDE_SET_SCHEMA = ["onnxruntime"]


class TritonModel:
    def __init__(
        self,
        *,
        versions: list[TritonModelVersion | dict] | None = None,
        backend: str | None = None,
        platform: str | None = None,
        max_batch_size: int | None = None,
        inputs: list[TritonDType | dict] | None = None,
        outputs: list[TritonDType | dict] | None = None,
        config: str | None = None,
        config_file: str | None = None,
        python_version: str | None = None,
        requirements_file: str | None = None,
        instance_group: list[dict] | None = None,
        test_cases: list[TestCase | dict] | None = None,
    ):
        """A Triton model.

        Args:
            versions: A list of TritonModelVersion objects representing the versions of the model.
            backend: The backend to use for the model, if provided. If not provided, it will be inferred from the model versions.
            config: Inline Config for Triton Model configuration, if provided.
            config_file: The path to the model's Triton configuration file (i.e. models/<model_name>/config.pbtxt), if provided.
            python_version: The python version to use for the model, if provided.
            requirements_file: The path to the model's requirements file (i.e. models/<model_name>/requirements.txt), if provided.
        """
        self.versions = versions or []
        self.versions = [TritonModelVersion(**mv) if isinstance(mv, dict) else mv for mv in self.versions]
        self.python_version = python_version
        self.requirements_file = requirements_file
        self.test_cases = test_cases or []
        self.test_cases = [
            TestCase(**test_case) if isinstance(test_case, dict) else test_case for test_case in self.test_cases
        ]
        self._check_version_numbers_all_or_none()

        self._check_config_and_config_file_isolated(config, config_file)
        if config:
            self.triton_config = parse_pbtxt(content=config) if config else {}
            self.triton_config_text = config if config else ""
        elif config_file:
            self.triton_config = parse_pbtxt(file_path=config_file) if config_file else {}
            self.triton_config_text = Path(config_file).read_text() if config_file else ""
        else:
            self.triton_config = {}
            self.triton_config_text = ""

        self._check_config_dups(self.triton_config, backend, platform, inputs, outputs, max_batch_size, instance_group)
        self.backend = backend or self.triton_config.get("backend", None)
        self.platform = platform or self.triton_config.get("platform", None)
        if inputs:
            inputs = [v if isinstance(v, TritonDType) else TritonDType(**v) for v in inputs]
            self.inputs = [{"name": v.name, "dtype": v.dtype, "dims": v.dims} for v in inputs]
        else:
            self.inputs = self.triton_config.get("input", None)
        if outputs:
            outputs = [v if isinstance(v, TritonDType) else TritonDType(**v) for v in outputs]
            self.outputs = [{"name": v.name, "dtype": v.dtype, "dims": v.dims} for v in outputs]
        else:
            self.outputs = self.triton_config.get("output", None)
        self.max_batch_size = max_batch_size or self.triton_config.get("max_batch_size", None)
        self.instance_group = instance_group or self.triton_config.get("instance_group", None)

        self._check_no_backend_duplicates()

        self.initialized: bool = False
        self.name: str | None = None
        self.path: Path | None = None
        self.flavor: str | None = None
        self.mlflow_model_paths: list[Path] | None = None
        self.mlflow_model_schema: tuple[int, dict, dict] | None = None
        self.ensemble_models: list[tuple[str, int]] | None = None
        self.triton_version: str | None = None
        self.decoupled_transaction_policy: bool | None = None

    @staticmethod
    def _check_config_and_config_file_isolated(config, config_file):
        if config and config_file:
            raise ValueError("Only one of config or config_file should be provided")

    @staticmethod
    def _check_config_dups(
        triton_config: dict,
        backend: str | None,
        platform: str | None,
        inputs: Any | None,
        outputs: Any | None,
        max_batch_size: int | None,
        instance_group: Any | None = None,
    ):
        if "backend" in triton_config and backend:
            raise ValueError("Backend is already specified in the Triton config")
        if "platform" in triton_config and platform:
            raise ValueError("Platform is already specified in the Triton config")
        if inputs and "input" in triton_config:
            raise ValueError("Input is already specified in the Triton config")
        if outputs and "output" in triton_config:
            raise ValueError("Output is already specified in the Triton config")
        if max_batch_size and "max_batch_size" in triton_config:
            raise ValueError("Max batch size is already specified in the Triton config")
        if instance_group and "instance_group" in triton_config:
            raise ValueError("Instance group is already specified in the Triton config")

    def _check_no_backend_duplicates(self):
        if self.backend and self.platform:
            raise ValueError("Both backend and platform are specified, which is not allowed. Please specify only one.")

    def _check_version_numbers_all_or_none(self):
        """Check that all versions have a version number or none of them have a version number"""
        if self.versions:
            version_numbers = [v.version for v in self.versions]
            if not all(version_numbers) and any(version_numbers):
                raise ValueError("All versions must have a version number or none of them should have a version number")

    def init(
        self,
        name: str,
        repo_path: Path,
        triton_version: str,
    ):
        click.secho(f"Initializing {name}")
        self.name = name
        self.path = repo_path.joinpath(self.name)
        self.triton_version = triton_version

        if "name" in self.triton_config and self.triton_config["name"] != self.name:
            raise ValueError(
                f"Model name in Triton config ({self.triton_config['name']}) does not match model name ({self.name})"
            )

        mlflow_artifact_version_uris = []
        for model_version in self.versions:
            if model_version.artifact_uri and model_version.artifact_uri.startswith("models:/"):
                mlflow_artifact_version_uris.append(model_version.artifact_uri)

        self.mlflow_model_paths = list()
        for artifact_uri in mlflow_artifact_version_uris:
            self.mlflow_model_paths.append(download_mlflow_model(artifact_uri))

        mlflow_model_schemas = list()
        for mlflow_model_path in self.mlflow_model_paths:
            with open(Path(mlflow_model_path).joinpath("MLmodel")) as f:
                mlflow_config = yaml.safe_load(f)
            try:
                mlflow_model_schemas.append(parse_mlflow_signature(mlflow_config))
            except (KeyError, ValueError):
                pass

        if mlflow_model_schemas:
            self.mlflow_model_schema = mlflow_model_schemas[0]
            for schema in mlflow_model_schemas[1:]:
                if schema != self.mlflow_model_schema:
                    raise ValueError(
                        f"MLflow model schemas do not match across versions, found: {self.mlflow_model_schema} and {schema}"
                    )

        self.flavor = get_flavor_for_models(mlflow_artifact_version_uris) if mlflow_artifact_version_uris else None
        flavor_backend_map = {  # map mlflow flavors to triton backends
            "python": "mlflow",
            "pytorch": "pytorch",
            "sentence_transformers": "mlflow",
            "transformers": "mlflow",
            "tensorflow": "tensorflow",
            "onnx": "onnxruntime",
        }
        mlflow_implied_backend = flavor_backend_map.get(self.flavor, None)
        if mlflow_artifact_version_uris and mlflow_implied_backend is None:
            raise ValueError(f"Unsupported MLflow flavor: {self.flavor}")

        if not self.backend and not self.platform:
            self.backend = mlflow_implied_backend

        if not self.backend and not self.platform:
            raise ValueError(
                "Backend must be specified or mlflow artifact versions must be specified\n"
                "Please specify a model backend via triton config or by passing the backend directly."
            )

        if self.python_version and self.backend not in {"python", "mlflow"}:
            raise ValueError("python_version can only be specified for python models")

        if not self.python_version and self.backend in {"python", "mlflow"} and mlflow_artifact_version_uris:
            self.python_version = get_python_version_for_mlflow_models(mlflow_artifact_version_uris)
        elif not self.python_version and self.backend in {"python", "mlflow"}:
            self.python_version = python_version_for_triton_version(self.triton_version)

        if self.python_version:
            self.python_version = parse_python_version(self.python_version)

        if self.requirements_file and self.backend not in {"python", "mlflow"}:
            raise ValueError("requirements_file can only be specified for python models")

        if not self.requirements_file and self.backend in {"python", "mlflow"} and mlflow_artifact_version_uris:
            self.requirements_file = get_mlflow_model_requirements(mlflow_artifact_version_uris)

        if self.backend == "pytorch" and self.mlflow_model_schema:
            mlflow_batch_size, mlflow_inputs, mlflow_outputs = self.mlflow_model_schema
            if not self.max_batch_size:
                self.max_batch_size = 1000 if mlflow_batch_size == -1 else mlflow_batch_size
            if not self.inputs:
                self.inputs = [
                    {"name": name, "dtype": dtype, "dims": dims} for name, (dtype, dims, _) in mlflow_inputs.items()
                ]
            if not self.outputs:
                self.outputs = [
                    {"name": name, "dtype": dtype, "dims": dims} for name, (dtype, dims, _) in mlflow_outputs.items()
                ]

        if self.backend == "pytorch" and not self.inputs:
            raise ValueError("inputs must be specified for PyTorch models")

        if self.backend == "pytorch" and not self.outputs:
            raise ValueError("outputs must be specified for PyTorch models")

        self.decoupled_transaction_policy = self.triton_config.get("model_transaction_policy", {}).get(
            "decoupled", False
        )

        default_version = 1
        for model_version in self.versions:
            if model_version.version:
                this_version = model_version.version
            elif model_version.artifact_uri and model_version.artifact_uri.startswith("models:/"):
                _models = find_matching_dbx_models(model_version.artifact_uri)
                if model_version:
                    this_version = int(_models[0].version)
                else:
                    raise ValueError(f"Could not find model version for {model_version.artifact_uri}")
            else:
                this_version = default_version
            version_dir = self.path.joinpath(str(this_version))
            model_version.version = this_version
            model_version.init(
                self.name,
                version_dir,
                self.backend,
                self.platform,
                self.flavor,
                self.test_cases,
                self.decoupled_transaction_policy,
            )
            default_version = this_version + 1

        self.versions = sorted(self.versions, key=lambda mv: mv.version)

        if self.platform == "ensemble":
            if len(self.versions) > 0:
                raise ValueError("Ensemble models cannot have versions specified")
            ensemble_version = TritonModelVersion(version=1)
            ensemble_version.init(
                self.name,
                self.path.joinpath("1"),
                self.backend,
                self.platform,
                self.flavor,
                self.test_cases,
                self.decoupled_transaction_policy,
            )
            self.versions.append(ensemble_version)
            self.ensemble_models = [
                (step["model_name"], step["model_version"]) for step in self.triton_config["ensemble_scheduling"]["step"]
            ]

        self.initialized = True

    def build(self, prep_conda_env: bool = True, prep_triton_stub: bool = True) -> None:
        """
        This method takes the initialized model container and handles processing and structuring the model artifacts.
        """
        self.path.mkdir(parents=True)

        if self.backend == "mlflow":
            assert not self.inputs, "'input' should not be specified in Triton config for MLflow models"
            assert not self.outputs, "'output' should not be specified in Triton config for MLflow models"
            mlflow_triton_config = build_config(
                self.mlflow_model_paths[0], default_max_batch_size=self.triton_config.get("max_batch_size", 1000)
            )
            self.triton_config_text += "\n" + mlflow_triton_config
            self.triton_config = parse_pbtxt(content=self.triton_config_text)

        if self.backend and "backend" not in self.triton_config:
            self.triton_config_text += f'\nbackend: "{self.backend}"'
            self.triton_config = parse_pbtxt(content=self.triton_config_text)

        if self.platform and "platform" not in self.triton_config:
            self.triton_config_text += f'\nplatform: "{self.platform}"'
            self.triton_config = parse_pbtxt(content=self.triton_config_text)

        if "version_policy" not in self.triton_config:
            self.triton_config_text += (
                f"\nversion_policy: {{ specific: {{ versions: {[mv.version for mv in self.versions]}}}}}"
            )
            self.triton_config = parse_pbtxt(content=self.triton_config_text)

        if "max_batch_size" not in self.triton_config and self.max_batch_size is not None:
            self.triton_config_text += f"\nmax_batch_size: {self.max_batch_size}"
            self.triton_config = parse_pbtxt(content=self.triton_config_text)

        if "instance_group" not in self.triton_config and self.instance_group is not None:
            instance_group_str = "\n".join(
                [
                    "\t{\n"
                    + "\n".join([f'\t\t{k}: {v if isinstance(v, int) else f"""{v}"""}' for k, v in ig.items()])
                    + "\n\t}"
                    for ig in self.instance_group
                ]
            )
            self.triton_config_text += f"\ninstance_group: [\n{instance_group_str}\n]"
            self.triton_config = parse_pbtxt(content=self.triton_config_text)

        if "input" not in self.triton_config and self.inputs:
            for triton_input in self.inputs:
                name, dtype, dims = triton_input["name"], triton_input["dtype"], triton_input["dims"]
                self.triton_config_text += (
                    f'\ninput [\n\t{{\n\t\tname: "{name}"\n\t\tdata_type: TYPE_{dtype}\n\t\tdims: {dims}\n\t}}\n]'
                )
            self.triton_config = parse_pbtxt(content=self.triton_config_text)

        if "output" not in self.triton_config and self.outputs:
            for triton_output in self.outputs:
                name, dtype, dims = triton_output["name"], triton_output["dtype"], triton_output["dims"]
                self.triton_config_text += (
                    f'\noutput [\n\t{{\n\t\tname: "{name}"\n\t\tdata_type: TYPE_{dtype}\n\t\tdims: {dims}\n\t}}\n]'
                )
            self.triton_config = parse_pbtxt(content=self.triton_config_text)

        if self.requirements_file and prep_conda_env:
            conda_pack_path, _, conda_pack_info = build_conda_env(
                self.python_version,
                Path(self.requirements_file),
            )
            dst_path = self.path.joinpath(conda_pack_path.name)
            click.secho(f"Creating hard link: {dst_path} -> {conda_pack_path}", fg="green")
            # rel_path = os.path.relpath(conda_pack_path, dst_path.parent)
            link_or_copy(conda_pack_path, dst_path)

            # point to python execution env, if it exists
            self.triton_config_text += f'\nparameters: {{key: "EXECUTION_ENV_PATH" value: {{string_value: "$$TRITON_MODEL_DIRECTORY/{conda_pack_path.name}"}}}}\n'

        if (
            self.python_version
            and self.python_version != python_version_for_triton_version(self.triton_version)
            and prep_triton_stub
        ):
            stub_path = build_triton_stub(self.python_version, f"r{self.triton_version}")
            dst_path = self.path.joinpath("triton_python_backend_stub")
            click.secho(f"Copying Triton Stub: {dst_path} -> {stub_path}", fg="green")
            shutil.copy2(stub_path, dst_path)

        for model_version in self.versions:
            click.secho(f"processing version: {model_version.version}", fg="green")
            model_version.build()

        self.path.joinpath("config.pbtxt").write_text(self.triton_config_text)

    def create_test_plan(self) -> TritonModelTestPlan:
        """Create a test plan for this model version."""
        if not self.initialized:
            raise ValueError("Model version must be initialized before creating a test plan")
        return TritonModelTestPlan(
            version_plans=[version.create_test_plan() for version in self.versions], ensemble_models=self.ensemble_models
        )



import os
from pathlib import Path
from typing import Tuple, Union

import boto3
import click

from tsbk import TSBK_DIR
from tsbk.utils import link_or_copy

s3r = boto3.resource("s3")
s3c = boto3.client("s3")


def split_s3_path(s3_path: str) -> Tuple[str, str]:
    """Splits an S3 URI into its bucket and key parts"""
    if not s3_path.startswith("s3://"):
        raise ValueError(f"Invalid s3 path: {s3_path}")

    path_parts = s3_path.replace("s3://", "").split("/")
    bucket = path_parts.pop(0)
    key = "/".join(path_parts)
    return bucket, key


def s3_path_exists(s3_path: str) -> bool:
    bucket, key = split_s3_path(s3_path)
    res = s3c.list_objects(Bucket=bucket, Prefix=key)
    if "Contents" in res:
        return True
    return False


def compute_cache_path(s3_path: str) -> Path:
    """Helper function for computing a local path for a cached object.
    This is done by joining the s3 bucket and key with the ARTIFACT_CACHE_DIR

    Args:
        s3_path: the path of a s3 object

    Returns:
        a local path that can be used to cache the object
    """
    bucket, key = split_s3_path(s3_path)
    info = s3c.head_object(Bucket=bucket, Key=key)
    etag = info["ETag"].strip('"')
    return TSBK_DIR / "cache" / "s3" / bucket / f"{key}.{etag}"


def download_s3_file(origin_path: str, dst_path: Union[str, Path]):
    """Downloads a file from s3 to a cache path and creates a symlink to this path at the dst_path for Trition to use when serving models that use this S3 artifact

    Args:
        origin_path: the path of the file to download ex: s3://my-bucket/my.file
        dst_path: the local path to download the file to
    """
    bucket, key = split_s3_path(origin_path)
    dst_path = Path(dst_path)
    dst_path.parent.mkdir(parents=True, exist_ok=True)
    local_cache_path = compute_cache_path(origin_path)
    if not local_cache_path.exists():
        local_cache_path.parent.mkdir(parents=True, exist_ok=True)
        click.secho(f"Downloading: s3://{bucket}/{key} -> {local_cache_path} (cache miss)", fg="blue")
        s3c.download_file(bucket, key, local_cache_path)
    click.secho(f"Creating hard link: {dst_path} -> {local_cache_path}", fg="blue")
    # rel_path = os.path.relpath(local_cache_path, dst_path.parent)
    link_or_copy(local_cache_path, dst_path)


def download_s3_path(origin_path: str, dst_path: Union[str, Path], **kwargs):
    """Downloads either a file or entire folder from s3 to a local caching path and creates symlinks in dst_path for each of these artifacts

    Args:
        origin_path: an S3 URI to either a file or folder
        dst_path: the path to download the file/folder to
    """

    bucket_name, prefix = split_s3_path(origin_path)
    bucket = s3r.Bucket(bucket_name)

    objects = list(bucket.objects.filter(Prefix=prefix))
    if len(objects) == 0:
        raise ValueError(f"No objects found at {origin_path}")

    for obj in bucket.objects.filter(Prefix=prefix):
        if obj.key.endswith("/"):  # Skip directories
            continue

        # Construct the local file path
        obj_local_path = os.path.join(dst_path, os.path.relpath(obj.key, prefix))

        # Download the file
        download_s3_file(f"s3://{bucket_name}/{obj.key}", obj_local_path)


def upload_s3_file(local_path: Union[str, Path], s3_path: str):
    """Upload a file to an S3 bucket

    Args:
        local_path: File to upload
        s3_path: the s3 location to copy the file to
    """
    bucket, key = split_s3_path(s3_path)
    click.secho(f"Uploading: {local_path} -> s3://{bucket}/{key}", fg="blue")
    s3c.upload_file(local_path, bucket, key)



from pydantic import BaseModel, ConfigDict

from tsbk import DEFAULT_TRITON_VERSION


class NumpyArraySpec(BaseModel):
    model_config = ConfigDict(extra="forbid")

    dtype: str
    """The data type of the NumPy array, which can be a string representation of the NumPy data type (e.g., 'float32', 'int64')."""
    dims: list[int]
    """The dimensions of the NumPy array, which is a list of integers representing the shape of the array."""
    data: list
    """The data of the NumPy array, which is a list of values that can be converted to the specified data type."""


class TestCaseSpec(BaseModel):
    model_config = ConfigDict(extra="forbid")

    inputs: dict[str, NumpyArraySpec]
    """The input data for the test case, which is a dictionary mapping input names to lists of values."""
    expected_outputs: dict[str, NumpyArraySpec]
    """The expected output data for the test case, which is a dictionary mapping output names to lists of values."""
    allow_nan: bool = False
    """Whether to allow NaN values in the output. If True, NaN values will not raise an error during validation."""
    allow_inf: bool = False
    """Whether to allow Inf values in the output. If True, Inf values will not raise an error during validation."""
    rtol: float = 1.0e-5
    """The relative tolerance to use for comparing floating point values. This is used to determine if two floating point values are close enough to be considered equal."""
    atol: float = 1.0e-7
    """The absolute tolerance to use for comparing floating point values. This is used to determine if two floating point values are close enough to be considered equal."""


class TritonDTypeSpec(BaseModel):
    model_config = ConfigDict(extra="forbid")

    name: str
    """The name of the Triton data type, which can be used to specify the input"""
    dtype: str
    """The Triton data type, which can be one of the supported Triton data types."""
    dims: list[int]
    """The dimensions of the Triton data type, which can be used to specify the shape of the input or output tensors."""


class TritonModelVersionSpec(BaseModel):
    model_config = ConfigDict(extra="forbid")

    artifact_uri: str | None = None
    """The URI of the model artifact, which can be an MLflow model or an S3 object."""
    python_model_file: str | None = None
    """The path to the Python model file, which is required for Python models."""
    version: int | None = None
    """The version number of the model"""
    test_cases: list[TestCaseSpec] | None = None
    """A list of test cases for the model version, which can be used to validate the model's behavior."""


class TritonModelSpec(BaseModel):
    model_config = ConfigDict(extra="forbid")

    versions: list[TritonModelVersionSpec] | None = None
    """A dictionary of model versions to deploy (optional, not required for ensemble models)"""
    backend: str | None = None
    """The Triton backend for this model to use, e.g. 'python', 'tensorrt', 'onnxruntime', etc."""
    platform: str | None = None
    """The tritonserver platform for this model to use, e.g. 'ensemble'"""
    max_batch_size: int | None = None
    """The maximum batch size for the model, which can be used to control the batching behavior"""
    inputs: list[TritonDTypeSpec] | None = None
    """A dictionary of input names to Triton data types, which can be used to specify the input tensors for the model."""
    outputs: list[TritonDTypeSpec] | None = None
    """A dictionary of output names to Triton data types, which can be used to specify the output tensors for the model."""
    config: str | None = None
    """Inline Config for Triton Model configuration"""
    config_file: str | None = None
    """ The path to the model's Triton configuration file (i.e. models/<model_name>/config.pbtxt)"""
    instance_group: list[dict] | None = None
    """The instance group configuration for the model, which can be used to specify the number of model instances and their placement."""
    python_version: str | None = None
    """ The python version to use for the model. """
    requirements_file: str | None = None
    """ The path to the model's requirements file. (i.e. models/<model_name>/requirements.txt)"""
    test_cases: list[TestCaseSpec] | None = None
    """A list of test cases for the model, which can be used to validate the model's behavior. Test cases supplied for a model are run for all model version."""


class TritonModelRepoSpec(BaseModel):
    model_config = ConfigDict(extra="forbid")

    name: str
    """The name of the Triton model repository, which is used as the directory name in the Triton model repository."""
    models: dict[str, TritonModelSpec]
    """A list of Triton models to deploy in the model repository."""
    triton_image: str = "nvcr.io/nvidia/tritonserver"
    """The Docker image to use for the Triton server."""
    triton_image_tag: str = f"{DEFAULT_TRITON_VERSION}-py3"
    """The Docker image tag to use for the Triton server."""



import json
import os
import shutil
from pathlib import Path
from typing import Optional

import click
import mlflow
import numpy as np
import pandas as pd
import yaml
from mlflow.entities.model_registry import ModelVersion as MlflowModelVersion
from mlflow.tracking import MlflowClient

from tsbk import TSBK_DIR
from tsbk.utils import link_or_copy, parse_requirements_file


def compute_cache_path(models_uri: str) -> Path:
    """
    Outputs a unique path for caching the specified dbx model locally.

    Args:
        models_uri: the models uri to cache

    Returns:
        provided cache_dir with the model workspace, name, and version/tag appended
    """
    if not models_uri.startswith("models:/"):
        raise ValueError(f"{models_uri} is not a valid model URI")
    model_dir = models_uri.split("models:/")[-1]
    return TSBK_DIR.joinpath("cache/mlflow", model_dir)


def get_model_uri_parts(model_uri: str) -> tuple[str, str, str]:
    """Returns the catalog, schema, and model name from the models_uri"""
    if "@" in model_uri:
        scheme, model_path = model_uri.split("/")
        model_name, model_id = model_path.split("@")
        return scheme, model_name, model_id
    else:
        scheme, model_name, model_id = model_uri.split("/")
        return scheme, model_name, model_id


def set_registry_uri_from_models_uri(models_uri: str):
    """Sets the correct registry URI for the model. This is determined by checking if the model has a pattern like:
    models:/{catalog}.{schema_name}.{model_name}/{version}. This version corresponds to the unity catalog, otherwise it's
    the normal model registry."""
    _, model_name, _ = get_model_uri_parts(models_uri)
    try:
        _, _, _ = model_name.split(".")
        mlflow.set_registry_uri(f"databricks-uc")
    except ValueError:
        mlflow.set_registry_uri(None)  # noqa


def _download_model_by_run(model_uri: str, dst_path: str):
    """Downloads an MLFLow model from a URI"""
    # set_registry_uri_from_models_uri(model_uri)
    client = MlflowClient()

    models = find_matching_dbx_models(model_uri)

    if len(models) == 0:
        raise ValueError(f"Model {model_uri} not found")

    [model] = models
    path = model.source.split("/")[-1]

    model_path = client.download_artifacts(model.run_id, path=path)
    file_names = os.listdir(model_path)

    for file_name in file_names:
        shutil.move(os.path.join(model_path, file_name), dst_path)
    shutil.rmtree(model_path)


def _download_model_by_model_uri(model_uri: str, dst_path: str):
    """Downloads an MLFLow model from a URI"""
    # set_registry_uri_from_models_uri(model_uri)
    mlflow.artifacts.download_artifacts(model_uri, dst_path=dst_path)


def _download_model_from_dbx(model_uri: str, dst_path: str):
    try:
        _download_model_by_model_uri(model_uri, dst_path)
    except Exception as e1:
        click.secho(f"Failed to download model by model uri from databricks: {e1}", fg="red")
        click.secho("Trying to download model by run", fg="yellow")
        try:
            _download_model_by_run(model_uri, dst_path)
        except Exception as e2:
            click.secho(f"Failed to download model by run from databricks: {e2}", fg="red")
            shutil.rmtree(dst_path, ignore_errors=True)
            raise e1


def download_mlflow_model(
    origin_path: str,
    dst_path: Optional[str] = None,
) -> Path:
    """Downloads an MLFLow model to a specific directory in the ARTIFACT_CACHE_DIR where it can be used in Symlinks in Triton model directories. Symlinks are used to allow various Triton model directories to share the same model artifact.

    Args:
        origin_path: the model to download in mlflow model format - models:/<your-model>/<your-model-version>
        dst_path: Path where a symlink to the cached model artifact will be created for Triton to use when serving the model. If not provided, the model will be saved at a designated path in the ARTIFACT_CACHE_DIR and no symlink to this path will be created.

    Returns:
        The path to the model
    """
    # we can't use symlinking on DBX cluster, so if a dst_path is provided, we download the model directly to that path
    local_cache_path = compute_cache_path(origin_path)
    if not local_cache_path.exists() or not list(local_cache_path.glob("*")):
        local_cache_path.parent.mkdir(parents=True, exist_ok=True)
        click.secho(f"Downloading: {origin_path} -> {local_cache_path} (cache miss)", fg="blue")
        _download_model_from_dbx(origin_path, local_cache_path.as_posix())
    if dst_path is not None:
        click.secho(f"Creating hard link: {dst_path} -> {local_cache_path}", fg="blue")
        shutil.copytree(local_cache_path, dst_path, copy_function=link_or_copy, dirs_exist_ok=True)
        return Path(dst_path)
    else:
        return local_cache_path


def get_mlflow_model_requirements(model_uris: list[str]) -> Path:
    """
    Pulls artifacts for any MLFlow models provided in configuration and verifies that all requirements files are matching.

    Args:
        model_uris: List of model_uris to check requirements for.
    Returns:
        Path to requirements file, if MLFlow model requirements are matching.
    """
    models_paths = []
    for uri in model_uris:
        path_to_model = download_mlflow_model(uri, dst_path=None)
        models_paths.append(path_to_model)

    requirements_paths = [p.joinpath("requirements.txt") for p in models_paths]
    model_requirements = [parse_requirements_file(rp) for rp in requirements_paths]
    # raise exception if there is a mismatch across requirements for model versions
    if set.union(*model_requirements) != set.intersection(*model_requirements):
        raise ValueError(
            f"MLFlow model versions have mismatching requirements. models requirements: {model_requirements}"
        )
    return requirements_paths[-1]


def get_python_version_for_mlflow_models(model_uris: list[str]) -> str:
    """Checks the python version of all the mlflow models and returns the python version to use
    Args:
        model_uris: List of model_uris to check requirements for.
    Returns:
        the python version to use
    """
    models_paths = []
    for uri in model_uris:
        path_to_model = download_mlflow_model(uri, dst_path=None)
        models_paths.append(path_to_model)

    python_versions = set()
    for p in models_paths:
        try:
            if p.joinpath("conda.yaml").exists():
                with open(p.joinpath("conda.yaml"), "r") as fp:
                    python_versions.add(yaml.safe_load(fp)["dependencies"][0].split("=")[1])
            else:
                with open(p.joinpath("python_env.yaml"), "r") as fp:
                    python_versions.add(yaml.safe_load(fp)["python"])
        except FileNotFoundError:
            click.secho("python_env.yaml not found, checking MLmodel", fg="yellow")
            with open(p.joinpath("MLmodel"), "r") as fp:
                mlflow_config = yaml.safe_load(fp)
                python_versions.add(mlflow_config["flavors"]["python_function"]["python_version"])

    if len(python_versions) > 1:
        raise ValueError(
            f"MLFlow model versions have mismatching python versions. models python versions: {python_versions}"
        )

    return list(python_versions)[0]


def find_matching_dbx_models(model_uri: str) -> list[MlflowModelVersion]:
    """Identifies models versions in the MLFlow model registry that match the provided model_uri"""
    # set_registry_uri_from_models_uri(model_uri)
    client = MlflowClient()
    _, name, identifier = get_model_uri_parts(model_uri)
    if "@" not in model_uri and identifier.isdigit():  # Static model with version
        try:
            return [client.get_model_version(name, identifier)]
        except mlflow.exceptions.RestException as e:
            if e.error_code == "RESOURCE_DOES_NOT_EXIST":
                return []
            raise e
    elif "@" not in model_uri and not identifier.isdigit():  # Model with stage name
        models = client.search_model_versions(filter_string=f"name = '{name}'")
        models = list(filter(lambda m: m.current_stage == identifier, models))
        return models
    elif "@" in model_uri:  # Model with alias
        try:
            return [client.get_model_version_by_alias(name, identifier)]
        except mlflow.exceptions.RestException as e:
            if e.error_code == "RESOURCE_DOES_NOT_EXIST":
                return []
            raise e

    else:
        raise ValueError(f"Invalid model uri: {model_uri}")


def get_flavor_for_models(model_uris: list[str]) -> str:
    """Checks the models metadata and returns the triton backend to use
    Args:
        model_uris: List of model_uris to check requirements for.
    Returns:
        the triton backend to use for the models
    """
    models_paths = []
    for uri in model_uris:
        path_to_model = download_mlflow_model(uri, dst_path=None)
        models_paths.append(path_to_model)

    model_flavors = set()
    for p in models_paths:
        with open(p.joinpath("MLmodel"), "r") as fp:
            mlflow_config = yaml.safe_load(fp)
            flavors = tuple(mlflow_config["flavors"].keys())
            model_flavors.add(flavors)

    if len(model_flavors) > 1:
        raise ValueError(f"MLFlow model versions have mismatching model flavors. model flavors: {model_flavors}")

    model_flavors = model_flavors.pop()
    flavors_list = ["pytorch", "tensorflow", "onnx", "transformers", "sentence_transformers"]
    for flavor in flavors_list:
        if flavor in model_flavors:
            return flavor
    else:
        return "python"


def get_input_example_from_model(model_uri: str) -> dict[str, np.ndarray] | None:
    model_path = download_mlflow_model(model_uri)
    mlflow_config = yaml.safe_load(Path(model_path).joinpath("MLmodel").read_text())
    example_info = mlflow_config.get("saved_input_example_info")
    if example_info is None:
        return None

    input_example_path = Path(model_path).joinpath(example_info["artifact_path"])
    match example_info["type"]:
        case "dataframe":
            df = pd.read_json(input_example_path, orient=example_info["pandas_orient"])
            table_values = df.to_dict(orient="list")
            if example_info["pandas_orient"] == "split":  # example is actually a dataframe
                example_values = {k: np.array(v) for k, v in table_values.items()}
            else:  # example is just a list of values, see mlflow_toupper_model test fixture
                example_values = {k: np.array(v[0]) for k, v in table_values.items()}
            return example_values
        case "ndarray":
            result = json.loads(input_example_path.read_text())
            if example_info.get("format") == "tf-serving":
                if type(result["inputs"]) == dict:
                    return {k: np.array(v) for k, v in result["inputs"].items()}
                else:
                    return {"input0": np.array(result["inputs"])}
            else:
                return {k: np.array(v) for k, v in result.items()}

        case "json_object":
            result = json.loads(input_example_path.read_text())
            return {"input0": np.array(result)}
        case _:
            raise ValueError(f"Unsupported input example type: {example_info['type']}")


# TODO determine if this is needed, as it is not used in the current codebase
"""
def match_model_inputs(
    triton_client: tritonclient.http.InferenceServerClient | tritonclient.grpc.InferenceServerClient,
    model_name: str,
    inputs: dict[str, np.ndarray],
    headers: dict | None = None,
) -> dict[str, np.ndarray]:
    ret_dict = dict()

    if isinstance(triton_client, tritonclient.http.InferenceServerClient):
        model_inputs = triton_client.get_model_config(model_name, headers=headers)["input"]
    else:
        model_inputs = triton_client.get_model_config(model_name, as_json=True)["config"]["input"]

    # TODO provide more advanced matching capabilities
    if len(model_inputs) == 1:
        if len(inputs) > 1:
            raise ValueError(f"Model {model_name} only has one input, but {len(inputs)} were provided")
        input_name = model_inputs[0]["name"]
        input_type = model_inputs[0]["data_type"].split("_")[1]
        input_type = "BYTES" if input_type == "STRING" else input_type
        ret_dict[input_name] = inputs.pop(list(inputs)[0]).astype(triton_to_np_dtype(input_type))
    else:
        for input_config in model_inputs:
            input_name = input_config["name"]
            input_type = input_config["data_type"].split("_")[1]
            input_type = "BYTES" if input_type == "STRING" else input_type
            if input_name in inputs:
                ret_dict[input_name] = inputs.pop(input_name).astype(triton_to_np_dtype(input_type))
            else:
                raise ValueError(f"Input '{input_name}' not found in provided inputs for model {model_name}")

    return ret_dict
"""



import numpy as np
from tritonclient.utils import np_to_triton_dtype


class TritonDType:
    TRITON_DTYPES = {
        "BOOL",
        "INT8",
        "INT16",
        "INT32",
        "INT64",
        "UINT8",
        "UINT16",
        "UINT32",
        "UINT64",
        "FP16",
        "FP32",
        "FP64",
        "STRING",
    }

    def __init__(self, name: str, dtype: np.dtype | str, dims: list[int]):
        self.name = name
        if dtype not in self.TRITON_DTYPES:
            if isinstance(dtype, str):
                dtype = np.dtype(dtype)

            self.dtype = np_to_triton_dtype(dtype)
            self.dtype = "STRING" if self.dtype == "BYTES" else self.dtype
        else:
            self.dtype = dtype

        assert (
            self.dtype in self.TRITON_DTYPES
        ), f"Invalid Triton dtype: {self.dtype}. Must be one of {self.TRITON_DTYPES}"
        self.dims = dims


class NumpyArray:
    def __init__(self, dtype: np.dtype | str, dims: list[int], data: list):
        if isinstance(dtype, str):
            dtype = np.dtype(dtype)
        self.dtype = dtype
        self.dims = dims
        self.data = data

    def to_numpy(self) -> np.ndarray:
        return np.array(self.data, dtype=self.dtype).reshape(self.dims)

    @classmethod
    def from_numpy(cls, array: np.ndarray) -> "NumpyArray":
        return cls(dtype=array.dtype, dims=list(array.shape), data=array.flatten().tolist())

    def to_dict(self) -> dict:
        return {
            "dtype": str(self.dtype),
            "dims": self.dims,
            "data": self.data,
        }



import queue
from collections import defaultdict
from functools import partial

import numpy as np
import tritonclient.grpc as grpcclient
from tritonclient.utils import np_to_triton_dtype, triton_to_np_dtype


class TritonGrpcModelClient:
    """A client for interacting with Triton models over HTTP"""

    def __init__(
        self,
        url: str,
        model_name: str,
        model_version: int | str | None = None,
        ca_certs: str | None = None,
        headers: dict | None = None,
    ):
        """A client for interacting with Triton models over HTTP

        Args:
            url: The url of the Triton server
            model_name: The name of the model to interact with
            model_version: The version of the model to interact with
            ca_certs: The path to the CA certificates file
            headers: Headers to include with each request
        """
        self.url = url
        self.model_name = model_name
        self.model_version = "" if model_version is None else str(model_version)

        self.client = self.create_client(url, ca_certs=ca_certs)
        self._model_config = None
        self.headers = headers or {}

    @staticmethod
    def create_client(url: str, ca_certs: str | None = None) -> grpcclient.InferenceServerClient:
        if url.startswith("http://"):
            ssl = False
            host = url[len("http://") :]
        elif url.startswith("https://"):
            ssl = True
            host = url[len("https://") :]
        else:
            raise ValueError(f"URL must start with 'http://' or 'https://', got: {url}")

        kwargs = {"url": host, "ssl": ssl}
        if ssl and ca_certs:
            kwargs["root_certificates"] = ca_certs
        return grpcclient.InferenceServerClient(**kwargs)

    def infer(
        self, inputs: dict[str, np.ndarray], output_names: list[str] | None = None, headers: dict[str, str] | None = None
    ) -> dict[str, np.ndarray]:
        """Run inference using the model

        Args:
            inputs: The inputs to the model
            output_names: The names of the outputs to return
            headers: The headers to use for the request

        Returns:
            The outputs of the model
        """
        headers = {**self.headers, **(headers or {})}
        self._validate_inputs(inputs)
        self._validate_output_names(output_names)

        infer_inputs = []
        for input_name in inputs:
            input_dtype = self.input_np_type(input_name)
            infer_inputs.append(
                grpcclient.InferInput(input_name, list(inputs[input_name].shape), np_to_triton_dtype(input_dtype))
            )
            infer_inputs[-1].set_data_from_numpy(inputs[input_name].astype(input_dtype))

        output_names = output_names or [output["name"] for output in self.model_outputs]
        infer_outputs = []
        for output_name in output_names:
            infer_outputs.append(grpcclient.InferRequestedOutput(output_name))

        response = self.client.infer(
            model_name=self.model_name,
            model_version=self.model_version,
            inputs=infer_inputs,
            outputs=infer_outputs,
            headers=headers,
        )

        return {name: response.as_numpy(name) for name in output_names}

    def infer_decoupled(
        self, inputs: dict[str, np.ndarray], output_names: list[str] | None = None, headers: dict[str, str] | None = None
    ) -> dict[str, np.ndarray]:
        class UserData:
            def __init__(self):
                self.completed_requests = queue.Queue()

        def callback(user_data, result, error):
            if error:
                user_data.completed_requests.put(error)
            else:
                user_data.completed_requests.put(result)

        user_data = UserData()
        self.client.start_stream(callback=partial(callback, user_data), headers=headers)  # noqa
        try:
            grpc_inputs = list()
            for name, arr in inputs.items():
                grpc_input = grpcclient.InferInput(name, list(arr.shape), np_to_triton_dtype(arr.dtype))
                grpc_input.set_data_from_numpy(arr)
                grpc_inputs.append(grpc_input)

            output_names = output_names or [output["name"] for output in self.model_outputs]
            infer_outputs = []
            for output_name in output_names:
                infer_outputs.append(grpcclient.InferRequestedOutput(output_name))

            self.client.async_stream_infer(
                model_name=self.model_name,
                model_version=self.model_version,
                inputs=grpc_inputs,
                enable_empty_final_response=True,
                request_id=str(1),
            )

            outputs = defaultdict(list)
            while True:
                data_item = user_data.completed_requests.get()
                if isinstance(data_item, grpcclient.InferenceServerException):
                    raise data_item
                if data_item.get_response(as_json=True)["parameters"]["triton_final_response"]["bool_param"]:
                    return {name: np.array(outputs[name]) for name in outputs}

                output_names = [output["name"] for output in data_item.get_response(as_json=True)["outputs"]]

                for output_name in output_names:
                    outputs[str(output_name)].append(data_item.as_numpy(output_name))
        finally:
            self.client.stop_stream()

    @property
    def model_config(self) -> dict:
        """The model configuration"""
        if self._model_config is None:
            self._model_config = self.client.get_model_config(self.model_name, headers=self.headers, as_json=True)[
                "config"
            ]
        return self._model_config

    @property
    def model_inputs(self) -> list[dict]:
        """The model input configuration"""
        return self.model_config["input"]

    @property
    def model_outputs(self) -> list[dict]:
        """The model output configuration"""
        return self.model_config["output"]

    def input_np_type(self, input_name: str):
        for model_input in self.model_inputs:
            if model_input["name"] == input_name:
                triton_dtype = model_input["data_type"].split("_")[1]
                if triton_dtype == "STRING":
                    triton_dtype = "BYTES"
                return triton_to_np_dtype(triton_dtype)
        raise ValueError(f"Input {input_name} not found in model inputs")

    def _validate_inputs(self, inputs: dict[str, np.ndarray]):
        for input_name in inputs:
            if not any(input_name == input["name"] for input in self.model_inputs):
                raise ValueError(f"Input {input_name} is not part of the model inputs")

    def _validate_output_names(self, output_names: list[str] | None):
        if output_names is None:
            return
        for output_name in output_names:
            if not any(output_name == output["name"] for output in self.model_outputs):
                raise ValueError(f"Output {output_name} is not part of the model outputs")



from dataclasses import dataclass

from tsbk.test_case import TestCase, TestResult
from tsbk.triton_grpc_model_client import TritonGrpcModelClient
from tsbk.triton_http_model_client import TritonHttpModelClient


@dataclass
class TritonModelVersionTestPlan:
    name: str
    version: int
    test_cases: list[TestCase]
    decoupled_transaction_policy: bool

    def run_tests(
        self, url: str, ca_certs: str | None = None, headers: dict | None = None, grpc: bool = False
    ) -> list[TestResult]:
        """Runs the test cases defined for this model version.

        Args:
            url: The URL of the Triton server to test against.
            ca_certs: Path to the CA certificates file for secure connections, if applicable.
            headers: Optional headers to include in the request.
            grpc: If True, use gRPC client, otherwise use HTTP client.

        Returns:
            A list of TestResult objects containing the results of the tests.
        """
        if not self.test_cases:
            return [
                TestResult(
                    success=False,
                    model_name=self.name,
                    model_version=self.version,
                    message=f"No test cases defined for {self.name}:{self.version}",
                    missing_test=True,
                )
            ]

        if self.decoupled_transaction_policy and not grpc:
            return [
                TestResult(
                    success=False,
                    model_name=self.name,
                    model_version=self.version,
                    message=f"Decoupled transaction policy is not supported for HTTP client. Use gRPC client for {self.name}:{self.version}",
                    decoupled_with_http=True,
                )
            ]

        kwargs = {
            "url": url,
            "headers": headers,
            "ca_certs": ca_certs,
            "model_name": self.name,
            "model_version": self.version,
        }

        model_client = TritonGrpcModelClient(**kwargs) if grpc else TritonHttpModelClient(**kwargs)

        results = list()
        for test_case in self.test_cases:
            print(f"Running {'grpc' if grpc else 'http'} test for {self.name}:{self.version} - ", end="")
            result = test_case.run(model_client, decoupled=self.decoupled_transaction_policy)
            print("✅" if result.success else "❌")
            results.append(result)
        return results

    def to_dict(self) -> dict:
        return {
            "name": self.name,
            "version": self.version,
            "test_cases": [tc.to_dict() for tc in self.test_cases],
            "decoupled_transaction_policy": self.decoupled_transaction_policy,
        }

    @classmethod
    def from_dict(cls, d: dict) -> "TritonModelVersionTestPlan":
        from tsbk.test_case import TestCase

        return cls(
            name=d["name"],
            version=d["version"],
            test_cases=[TestCase.from_dict(tc) for tc in d["test_cases"]],
            decoupled_transaction_policy=d["decoupled_transaction_policy"],
        )


@dataclass
class TritonModelTestPlan:
    version_plans: list[TritonModelVersionTestPlan]
    ensemble_models: list[tuple[str, int]] | None

    def run_tests(
        self, url: str, ca_certs: str | None = None, headers: dict | None = None, grpc: bool = False
    ) -> list[TestResult]:
        """Runs the test cases defined for this model version.

        Args:
            url: The URL of the Triton server to test against.
            ca_certs: Path to the CA certificates file for secure connections, if applicable.
            headers: Optional headers to include in the request.
            grpc: If True, use gRPC client, otherwise use HTTP client.

        Returns:
            A list of TestResult objects containing the results of the tests.
        """
        results = list()
        for version_plan in self.version_plans:
            results.extend(version_plan.run_tests(url, ca_certs=ca_certs, headers=headers, grpc=grpc))
        return results

    def to_dict(self) -> dict:
        return {
            "version_plans": [vp.to_dict() for vp in self.version_plans],
            "ensemble_models": self.ensemble_models,
        }

    @classmethod
    def from_dict(cls, d: dict) -> "TritonModelTestPlan":
        return cls(
            version_plans=[TritonModelVersionTestPlan.from_dict(vp) for vp in d["version_plans"]],
            ensemble_models=[tuple(model) for model in d["ensemble_models"] or []],  # noqa
        )


@dataclass
class TritonModelRepoTestPlan:
    model_plans: list[TritonModelTestPlan]
    repo_models: list[tuple[str, int]]

    def run_tests(
        self, url: str, ca_certs: str | None = None, headers: dict | None = None, grpc: bool = False
    ) -> list[TestResult]:
        """Runs the test cases defined for all models in this repository against the Triton server.

        Args:
            url: The URL of the Triton server to test against.
            ca_certs: Path to CA certificates for secure connections, if applicable.
            headers: Additional headers to include in the requests.
            grpc: Whether to use gRPC for testing.

        Returns:
            A list of TestResult objects containing the results of the tests.
        """
        self.check_server_matches(url, ca_certs=ca_certs, headers=headers, grpc=grpc)
        results: list[TestResult] = []
        for model in self.model_plans:
            results.extend(model.run_tests(url, ca_certs=ca_certs, headers=headers, grpc=grpc))

        for result in results:
            if result.missing_test and self._test_result_covered_by_ensemble(result):
                print(f"Testing {result.model_name}:{result.model_version} are covered by an ensemble model ✅")
                result.success = True
            elif result.missing_test:
                print(f"No test cases defined for {result.model_name}:{result.model_version} ❌")
            elif not result.success:
                print(f"Failure details for {result.model_name}/{result.model_version}")
                if result.message:
                    print(f"Message: {result.message}")
                if result.exception:
                    print(f"Exception: {result.exception}")
                if result.traceback:
                    print(f"Traceback: \n{result.traceback}")
                print("")
        return results

    def check_server_matches(
        self, url: str, ca_certs: str | None = None, headers: dict | None = None, grpc: bool = False
    ):
        """Checks that the server has the same models as the repo."""
        server_models = set(self.get_server_models(url, ca_certs=ca_certs, headers=headers, grpc=grpc))
        repo_models = set(self.repo_models)
        if server_models != repo_models:
            raise ValueError(
                f"Server models do not match repo models. Missing models: {repo_models - server_models}, Unexpected models: {server_models - repo_models}"
            )

    @staticmethod
    def get_server_models(
        url: str, ca_certs: str | None = None, headers: dict | None = None, grpc: bool = False
    ) -> list[tuple[str, int]]:
        """Returns a list of models that are currently loaded in the server"""
        if not grpc:
            triton_client = TritonHttpModelClient.create_client(url, ca_certs=ca_certs)
            model_index = triton_client.get_model_repository_index(headers=headers)
            return [(model["name"], int(model["version"])) for model in model_index]
        else:
            triton_client = TritonGrpcModelClient.create_client(url, ca_certs=ca_certs)
            model_index = triton_client.get_model_repository_index(headers=headers)
            return [(model.name, int(model.version)) for model in model_index.models]

    def _test_result_covered_by_ensemble(self, result: TestResult) -> bool:
        """Checks if a test result is covered by an ensemble model."""
        for model in self.model_plans:
            if model.ensemble_models:
                for model_name, model_version in model.ensemble_models:
                    if result.model_name == model_name and result.model_version == model_version:
                        return True
        return False

    def to_dict(self) -> dict:
        return {
            "model_plans": [mp.to_dict() for mp in self.model_plans],
            "repo_models": self.repo_models,
        }

    @classmethod
    def from_dict(cls, d: dict) -> "TritonModelRepoTestPlan":
        return cls(
            model_plans=[TritonModelTestPlan.from_dict(mp) for mp in d["model_plans"]],
            repo_models=[tuple(model) for model in d["repo_models"]],  # noqa
        )



import json
import sys

import click
import msgpack
import yaml

from tsbk import __version__
from tsbk.model_repo import TritonModelRepo
from tsbk.spec import TritonModelRepoSpec


@click.group()
def tsbk():
    pass


@tsbk.command(help="Print current tsbk version")
def version():
    click.echo(f"tsbk version: {__version__}")


@tsbk.command(help="Build a Tritonserver model repository from a config file")
@click.argument("config_file", type=click.Path(exists=True, dir_okay=False))
@click.argument("output_path", type=click.Path(exists=False))
def build(config_file, output_path):
    """Build a Tritonserver model repository from a config file."""
    with open(config_file, "r") as f:
        click.secho(f"Reading configuration from: {config_file}")
        config = yaml.safe_load(f)

    spec = TritonModelRepoSpec(**config)
    repo = TritonModelRepo(path=output_path, **spec.model_dump())
    repo.build()
    click.secho(f"Model repository built at: {repo.path}")


@tsbk.command(help="Run a Tritonserver model repository from a config file")
@click.argument("config_file", type=click.Path(exists=True, dir_okay=False))
@click.argument("repo_path", type=click.Path(exists=False))
@click.option("--http-port", default=8000, show_default=True, help="HTTP port for Triton server")
@click.option("--grpc-port", default=8001, show_default=True, help="gRPC port for Triton server")
@click.option("--metrics-port", default=8002, show_default=True, help="Metrics port for Triton server")
@click.option("--shm-size", default="4G", show_default=True, help="Shared memory size for Docker container")
@click.option("--gpus", is_flag=True, help="Enable GPU support")
@click.option("--env-file", type=click.Path(exists=True, dir_okay=False), default=None, help="Path to Docker env file")
@click.option("--detach", is_flag=True, help="Run container in detached mode")
@click.option("--test", is_flag=True, help="Run container in detached mode")
def run(config_file, repo_path, http_port, grpc_port, metrics_port, shm_size, gpus, env_file, detach, test):
    """Run a Tritonserver model repository from a config file."""
    if test:
        detach = True
    with open(config_file, "r") as f:
        click.secho(f"Reading configuration from: {config_file}")
        config = yaml.safe_load(f)
    spec = TritonModelRepoSpec(**config)
    repo = TritonModelRepo(path=repo_path, **spec.model_dump())
    repo.build()
    repo.run(
        http_port=http_port,
        grpc_port=grpc_port,
        metrics_port=metrics_port,
        shm_size=shm_size,
        gpus=gpus,
        env_file=env_file,
        detach=detach,
    )
    if detach:
        click.secho(f"Triton server started for repo at: {repo.path}")

    if test:
        click.secho("Running tests for the Triton server repository...")
        http_results = repo.test(url=repo.http_url)
        grpc_results = repo.test(url=repo.grpc_url, grpc=True)
        all_results = http_results + grpc_results
        success = all(r.success for r in all_results)
        repo.stop()
        sys.exit(0 if success else 1)


@tsbk.command(help="Test a Tritonserver model repository against a running server")
@click.argument("config_file", type=click.Path(exists=True, dir_okay=False))
@click.argument("repo_path", type=click.Path(exists=False))
@click.option("--url", required=True, help="URL of the running Triton server (http or grpc)")
@click.option(
    "--ca-certs",
    type=click.Path(exists=True, dir_okay=False),
    default=None,
    help="Path to CA certificates for secure connections",
)
@click.option("--headers", type=str, default=None, help="Additional headers as JSON string")
@click.option("--grpc", is_flag=True, help="Use gRPC for testing")
def test(config_file, repo_path, url, ca_certs, headers, grpc):
    """Test a Tritonserver model repository against a running server."""
    with open(config_file, "r") as f:
        click.secho(f"Reading configuration from: {config_file}")
        config = yaml.safe_load(f)
    spec = TritonModelRepoSpec(**config)
    repo = TritonModelRepo(path=repo_path, **spec.model_dump())
    headers_dict = json.loads(headers) if headers else None
    click.secho(f"Testing Triton server at: {url}", fg="blue")
    results = repo.test(url=url, ca_certs=ca_certs, headers=headers_dict, grpc=grpc)
    success = all(r.success for r in results)
    for r in results:
        click.secho(
            f"Test for model {r.model_name} (version {r.model_version}): {'PASS' if r.success else 'FAIL'}",
            fg="green" if r.success else "red",
        )
    sys.exit(0 if success else 1)


@tsbk.command(help="Create a test plan for a Tritonserver model repository and serialize it to a file")
@click.argument("config_file", type=click.Path(exists=True, dir_okay=False))
@click.argument("repo_path", type=click.Path(exists=False))
@click.argument("output_path", type=click.Path(exists=False))
def create_test_plan(config_file, repo_path, output_path):
    """Create a test plan and serialize it to a file."""
    with open(config_file, "r") as f:
        click.secho(f"Reading configuration from: {config_file}")
        config = yaml.safe_load(f)
    spec = TritonModelRepoSpec(**config)
    repo = TritonModelRepo(path=repo_path, **spec.model_dump())
    test_plan = repo.create_test_plan()
    with open(output_path, "wb") as f:
        f.write(msgpack.packb(test_plan.to_dict()))
    click.secho(f"Test plan created and saved to: {output_path}", fg="green")


@tsbk.command(help="Run a serialized test plan against a running Tritonserver")
@click.argument("test_plan_path", type=click.Path(exists=True, dir_okay=False))
@click.option("--url", required=True, help="URL of the running Triton server (http or grpc)")
@click.option(
    "--ca-certs",
    type=click.Path(exists=True, dir_okay=False),
    default=None,
    help="Path to CA certificates for secure connections",
)
@click.option("--headers", type=str, default=None, help="Additional headers as JSON string")
@click.option("--grpc", is_flag=True, help="Use gRPC for testing")
def run_test_plan(test_plan_path, url, ca_certs, headers, grpc):
    """Run a serialized test plan against a running Tritonserver."""
    import json

    from tsbk.testing import TritonModelRepoTestPlan

    with open(test_plan_path, "rb") as f:
        test_plan_dict = msgpack.unpackb(f.read(), raw=False)
        test_plan = TritonModelRepoTestPlan.from_dict(test_plan_dict)
    headers_dict = json.loads(headers) if headers else None
    click.secho(f"Running test plan against Triton server at: {url}", fg="blue")
    results = test_plan.run_tests(url=url, ca_certs=ca_certs, headers=headers_dict, grpc=grpc)
    success = all(r.success for r in results)
    for r in results:
        click.secho(
            f"Test for model {r.model_name} (version {r.model_version}): {'PASS' if r.success else 'FAIL'}",
            fg="green" if r.success else "red",
        )
    sys.exit(0 if success else 1)


if __name__ == "__main__":
    tsbk()



import os
from pathlib import Path

__version__ = "1.8.2"

# env vars
TSBK_DIR = Path(os.environ.get("TSBK_DIR", "./.tsbk"))
TSBK_S3_PREFIX = os.environ.get("TSBK_S3_PREFIX")
TSBK_K8S_SERVICE_ACCOUNT = os.environ.get("TSBK_K8S_SERVICE_ACCOUNT", "default")

# constants
DEFAULT_TRITON_VERSION = "25.08"

from .cli import tsbk
from .model import TritonModel
from .model_repo import TritonModelRepo
from .model_version import TritonModelVersion
from .test_case import TestCase
from .triton_grpc_model_client import TritonGrpcModelClient
from .triton_http_model_client import TritonHttpModelClient
from .types import TritonDType



import math
import re
from pathlib import Path
from typing import Optional, Tuple, Union


def parse_pbtxt(*, file_path: Union[str, Path] | None = None, content: str | None = None) -> dict:
    """Parses a triton pbtxt config file and returns the result as a dictionary

    Args:
        file_path: the path to the config file
        content: the content of the config file as a string

    Returns:
        configuration as a dictionary
    """
    assert file_path is not None or content is not None, "Either file_path or content must be provided"
    assert file_path is None or content is None, "Only one of file_path or content can be provided"

    if file_path:
        with open(file_path, "r") as f:
            content = f.readlines()
    else:
        content = content.splitlines()

    for i, line in enumerate(content):
        if "#" in line:
            new_line = line.split("#")[0]
            content[i] = new_line + "\n"

    content = "\n".join(content)

    sections = re.split(r"\n(?=\w)", content)
    parsed_data = {}

    for section in sections:
        data = parse_section(section.strip())
        if data is not None:
            key, value = data
            if key in parsed_data:
                if not isinstance(parsed_data[key], list):
                    parsed_data[key] = [parsed_data[key]]
                if isinstance(value, list):
                    parsed_data[key].extend(value)
                else:
                    parsed_data[key].append(value)
            else:
                parsed_data[key] = value

    return parsed_data


def parse_section(section: str) -> Optional[Tuple[str, Union[str, list, dict]]]:
    """Helper function for parsing section text"""
    if section.startswith("#") or section == "":
        return None

    s_idx, _ = first_sep_token(section)

    section_name, body = section[:s_idx], section[s_idx:].strip(":")
    return section_name.strip(), parse_section_body(body)


def parse_section_body(body: str):
    """Helper function for parsing section body data"""
    text = body.strip()

    if text[0] == "{" and text[-1] == "}":
        text = text[1:-1]
        result = {}
        while len(text) != 0:
            s_idx, token = first_sep_token(text)
            key, value = text[:s_idx].strip(",").strip(), text[s_idx:].strip(":").strip()
            if value[0] == "{":
                matching_idx = find_matching_brace(value, s_index=0)
                value, text = value[0 : matching_idx + 1], value[matching_idx + 1 :]
            elif value[0] == "[":
                matching_idx = find_matching_brace(value, s_index=0, l_char="[", r_char="]")
                value, text = value[0 : matching_idx + 1], value[matching_idx + 1 :]
            elif ":" in value:
                value, text = value.split(None, maxsplit=1)
            else:
                text = ""
            result[key.strip()] = parse_section_body(value)
        return result
    elif text[0] == "[" and text[-1] == "]" and "{" not in text:
        text = text[1:-1].strip()
        result = []
        if text:
            for t in text.split(","):
                result.append(parse_section_body(t))
        return result
    elif text[0] == "[" and text[-1] == "]" and "{" in text:
        text = text[1:-1].strip()
        result = []
        while len(text) != 0:
            matching_idx = find_matching_brace(text, s_index=0)
            result.append(parse_section_body(text[0 : matching_idx + 1]))
            text = text[matching_idx + 1 :].strip("\n\t ,")
        return result
    else:
        # parse literal
        try:
            return int(text.strip())
        except ValueError:
            return text.replace('"', "")


def first_sep_token(text: str, sep_tokens: tuple[str, ...] = (":", "{", "[")) -> tuple[int, str]:
    """Finds the first occurrence of any of the separator tokens in the text and returns its index and the token"""
    indices = [text.find(token) for token in sep_tokens if text.find(token) != -1]
    if not indices:
        raise ValueError(f'No separator tokens {sep_tokens} found in the text: "{text}"')
    min_index = min(indices)
    return min_index, text[min_index]


def find_matching_brace(text: str, s_index: int = 0, l_char: str = "{", r_char: str = "}"):
    """Finds the properly matching curl brace '}' that matches the brace specified by s_index"""
    if text[s_index] != l_char:
        raise ValueError("There is no curly brace at s_index in the provided text")
    counter = 0
    i = s_index + 1
    while i < len(text):
        if text[i] == r_char and counter == 0:
            return i
        elif text[i] == r_char:
            counter -= 1
        elif text[i] == l_char:
            counter += 1
        i += 1
    raise ValueError("There is no matching brace")



import hashlib
import json
from pathlib import Path

import click
from mlflow_backend_utils.sdk import build_conda_pack as _build_conda_pack

from tsbk import TSBK_DIR
from tsbk.utils import get_platform_desc

CONDA_BUILD_VERSION = 5


def calc_hash(
    requirements_file_bytes: bytes,
    platform_bytes: bytes,
    python_version: str,
    build_env: dict[str, str] = None,
    conda_extras: str = "",
) -> str:
    """Calculates the hash for the environment based on requirements and platform

    Args:
        requirements_file_bytes: byte representation of requirements file contents i.e. Path(path/to/requirements.txt).read_bytes()
        platform_bytes: bytes describing current platform (from import platform), platform.encode()
        python_version: python version for environment
        build_env: extra environment variables to set when running the conda build script
        conda_extras: extra conda packages to install

    Returns:
        hash containing conda env information used as file name for this env in model repo and s3 cache.
    """
    hash_str = (
        requirements_file_bytes
        + platform_bytes
        + python_version.encode()
        + str(CONDA_BUILD_VERSION).encode()
        + conda_extras.encode()
    )

    if build_env:
        hash_str += "".join([f"{k}_{v}" for k, v in sorted(build_env.items())]).encode()

    return hashlib.sha256(hash_str).hexdigest()


def build_conda_env(
    python_version: str,
    requirements_path: Path,
) -> tuple[Path, Path, dict]:
    """Builds a conda environment given the requirements and packages it up with conda-pack and writes
    the resulting archive to the output_dir

    Args:
        python_version: the python version to use for the environment
        requirements_path: The path to the requirements file to use

    Returns:
        path to the output archive
        path to json file with info for the environment
        dictionary of the conda environment info
    """
    # Hash inputs
    req_bytes = requirements_path.read_bytes()
    pf = get_platform_desc().encode()
    hsh = calc_hash(req_bytes, pf, python_version)

    env_data = {
        "python_version": python_version,
        "requirements": req_bytes.decode().split("\n"),
        "platform": pf.decode(),
        "hash": hsh,
        "build_version": CONDA_BUILD_VERSION,
    }

    output_dir = TSBK_DIR.joinpath("conda_packs")
    output_dir.mkdir(parents=True, exist_ok=True)
    output_path = output_dir.joinpath(f"{hsh}.tar.gz")
    info_path = output_dir.joinpath(f"{hsh}.json")

    if output_path.exists() and info_path.exists():
        return output_path, info_path, env_data

    click.secho(f"Running conda build script for {requirements_path} on platform {pf.decode()}", fg="blue")
    _build_conda_pack(
        python_version=python_version,
        requirements=requirements_path,
        output_path=output_path,
    )

    with open(info_path, "w") as fp:
        json.dump(env_data, fp)

    return output_path, info_path, env_data
